﻿using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using MVR.FileManagementSecure;
using SimpleJSON;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

namespace JaxZoa
{

    public class SimpleKeybind : MVRScript
    {
        protected JSONStorableFloat triggerCount;

        protected KeyBindTrigger currentTrigger;
        protected List<KeyBindTrigger> triggers;

        protected JSONStorableStringChooser triggerChooser;
        protected JSONStorableStringChooser keyChooser;
        protected UIDynamicTextField nameInputField;

        protected List<string> keyNames;
        protected List<KeyCode> keyCodes;

        // Use this for initialization
        public override void Init()
        {
            SimpleTriggerHandler.LoadAssets();

            keyNames = GetKeybindNames();
            keyCodes = GenereateTriggersKeyCodeList();

            triggers = new List<KeyBindTrigger>();

            activeSliders = new List<UIDynamicSlider>();
            activeToggles = new List<UIDynamicToggle>();
            activeButtons = new List<UIDynamicButton>();
            activePopups = new List<UIDynamicPopup>();
            activeTextFields = new List<UIDynamicTextField>();
            activeSpacers = new List<UIDynamic>();

            triggerCount = new JSONStorableFloat("TriggerCount", 0, 0, 1000, true, false);
            RegisterFloat(triggerCount);

            triggerChooser = new JSONStorableStringChooser("TriggerChooser", new List<string>(), "Select Trigger", "Triggers", Callback_OnTriggerSelected);
            triggerChooser.popupOpenCallback += Callback_GenerateTriggerIDList;
            RegisterStringChooser(triggerChooser);

            keyChooser = new JSONStorableStringChooser("KeyChooser", keyNames, "Select Key", "Keybind", Callback_OnKeySelected);
            RegisterStringChooser(keyChooser);

            CreateMainUI();
        }

        // Update is called once per frame
        void Update()
        {
            foreach(KeyBindTrigger _t in triggers)
            {
                if (_t.ShiftMod.val && !Input.GetKey(KeyCode.LeftShift) && !Input.GetKey(KeyCode.RightShift))
                    continue;

                if (_t.CtrlMod.val && !Input.GetKey(KeyCode.LeftControl) && !Input.GetKey(KeyCode.RightControl))
                    continue;

                if (_t.AltMod.val && !Input.GetKey(KeyCode.LeftAlt) && Input.GetKey(KeyCode.RightAlt))
                    continue;

                if (!_t.ShiftMod.val && (Input.GetKey(KeyCode.LeftShift) || Input.GetKey(KeyCode.RightShift)))
                    continue;

                if (!_t.CtrlMod.val && (Input.GetKey(KeyCode.LeftControl) || Input.GetKey(KeyCode.RightControl)))
                    continue;

                if (!_t.AltMod.val && (Input.GetKey(KeyCode.LeftAlt) || Input.GetKey(KeyCode.RightAlt)))
                    continue;



                if (!Input.GetKeyDown(_t.Key))
                    continue;

                _t.Trigger.Trigger();
            }
        }

        void Destroy()
        {
            foreach(KeyBindTrigger _t in triggers)
            {
                _t.OnDestroy();
            }
        }

        void CreateMainUI()
        {
            CreateFilterablePopup(triggerChooser, false);

            nameInputField = CreateTextInput(new JSONStorableString("nameInput", "Keybind name"), 50, false);

            UIDynamicButton _btn = CreateButton("Create New Keybind", false);
            _btn.button.onClick.AddListener(Callback_CreateNewTrigger);

            _btn = CreateButton("Delete Selected Keybind", false);
            _btn.button.onClick.AddListener(Callback_DestroyCurrentTrigger);

            _btn = CreateButton("Rename Selected Keybind", false);
            _btn.button.onClick.AddListener(Callback_RenameTrigger);
            

        }

        void UpdateKeyBindUI()
        {
            ClearUI();

            if (currentTrigger == null)
                return;

            var text = CreateTempStaticText("Title", currentTrigger.ActionName.val, 75);

            text.UItext.fontSize = 55;

            CreateTempButton("Edit Trigger", Callback_OpenTriggerPanel);

            keyChooser.val = currentTrigger.KeyID;

            CreateTempFilterablePopup(keyChooser);

            CreateTempToggle(currentTrigger.ShiftMod);
            CreateTempToggle(currentTrigger.CtrlMod);
            CreateTempToggle(currentTrigger.AltMod);
        }

        protected List<string> GetKeybindNames()
        {
            List<string> list = new List<string>();

            list.Add("Key_A");
            list.Add("Key_B");
            list.Add("Key_C");
            list.Add("Key_D");
            list.Add("Key_E");
            list.Add("Key_F");
            list.Add("Key_G");
            list.Add("Key_H");
            list.Add("Key_I");
            list.Add("Key_J");
            list.Add("Key_K");
            list.Add("Key_L");
            list.Add("Key_M");
            list.Add("Key_N");
            list.Add("Key_O");
            list.Add("Key_P");
            list.Add("Key_Q");
            list.Add("Key_R");
            list.Add("Key_S");
            list.Add("Key_T");
            list.Add("Key_U");
            list.Add("Key_V");
            list.Add("Key_W");
            list.Add("Key_X");
            list.Add("Key_Y");
            list.Add("Key_Z");          

            list.Add("Alpha0");
            list.Add("Alpha1");
            list.Add("Alpha2");
            list.Add("Alpha3");
            list.Add("Alpha4");
            list.Add("Alpha5");
            list.Add("Alpha6");
            list.Add("Alpha7");
            list.Add("Alpha8");
            list.Add("Alpha9");

            list.Add("Comma");
            list.Add("Period");
            list.Add("Backspace");

            list.Add("UpArrow");
            list.Add("DownArrow");
            list.Add("LeftArrow");
            list.Add("RightArrow");

            list.Add("Insert");
            list.Add("Delete");
            list.Add("Home");
            list.Add("End");
            list.Add("PageUp");
            list.Add("PageDown");

            list.Add("Num0");
            list.Add("Num1");
            list.Add("Num2");
            list.Add("Num3");
            list.Add("Num4");
            list.Add("Num5");
            list.Add("Num6");
            list.Add("Num7");
            list.Add("Num8");
            list.Add("Num9");

            list.Add("KeypadPeriod");
            list.Add("KeypadEnter");
            list.Add("KeypadEquals");
            list.Add("KeypadMultiply");
            list.Add("KeypadDivide");
            list.Add("KeypadPlus");
            list.Add("KeypadMinus");

            list.Add("F1");
            list.Add("F2");
            list.Add("F3");
            list.Add("F4");

            list.Add("F5");
            list.Add("F6");
            list.Add("F7");
            list.Add("F8");

            list.Add("F9");
            list.Add("F10");
            list.Add("F11");
            list.Add("F12");

            return list;
        }

        protected List<KeyCode> GenereateTriggersKeyCodeList()
        {
            List<KeyCode> _result = new List<KeyCode>();

            _result.Add(KeyCode.A);
            _result.Add(KeyCode.B);
            _result.Add(KeyCode.C);
            _result.Add(KeyCode.D);
            _result.Add(KeyCode.E);
            _result.Add(KeyCode.F);
            _result.Add(KeyCode.G);
            _result.Add(KeyCode.H);
            _result.Add(KeyCode.I);
            _result.Add(KeyCode.J);
            _result.Add(KeyCode.K);
            _result.Add(KeyCode.L);
            _result.Add(KeyCode.M);
            _result.Add(KeyCode.N);
            _result.Add(KeyCode.O);
            _result.Add(KeyCode.P);
            _result.Add(KeyCode.Q);
            _result.Add(KeyCode.R);
            _result.Add(KeyCode.S);
            _result.Add(KeyCode.T);
            _result.Add(KeyCode.U);
            _result.Add(KeyCode.V);
            _result.Add(KeyCode.W);
            _result.Add(KeyCode.X);
            _result.Add(KeyCode.Y);
            _result.Add(KeyCode.Z);

            _result.Add(KeyCode.Alpha0);
            _result.Add(KeyCode.Alpha1);
            _result.Add(KeyCode.Alpha2);
            _result.Add(KeyCode.Alpha3);
            _result.Add(KeyCode.Alpha4);
            _result.Add(KeyCode.Alpha5);
            _result.Add(KeyCode.Alpha6);
            _result.Add(KeyCode.Alpha7);
            _result.Add(KeyCode.Alpha8);
            _result.Add(KeyCode.Alpha9);

            _result.Add(KeyCode.Comma);
            _result.Add(KeyCode.Period);
            _result.Add(KeyCode.Backspace);

            _result.Add(KeyCode.UpArrow);
            _result.Add(KeyCode.DownArrow);
            _result.Add(KeyCode.LeftArrow);
            _result.Add(KeyCode.RightArrow);

            _result.Add(KeyCode.Insert);
            _result.Add(KeyCode.Delete);
            _result.Add(KeyCode.Home);
            _result.Add(KeyCode.End);
            _result.Add(KeyCode.PageUp);
            _result.Add(KeyCode.PageDown);

            _result.Add(KeyCode.Keypad0);
            _result.Add(KeyCode.Keypad1);
            _result.Add(KeyCode.Keypad2);
            _result.Add(KeyCode.Keypad3);
            _result.Add(KeyCode.Keypad4);
            _result.Add(KeyCode.Keypad5);
            _result.Add(KeyCode.Keypad6);
            _result.Add(KeyCode.Keypad7);
            _result.Add(KeyCode.Keypad8);
            _result.Add(KeyCode.Keypad9);

            _result.Add(KeyCode.KeypadPeriod);
            _result.Add(KeyCode.KeypadEnter);
            _result.Add(KeyCode.KeypadEquals);
            _result.Add(KeyCode.KeypadMultiply);
            _result.Add(KeyCode.KeypadDivide);
            _result.Add(KeyCode.KeypadPlus);
            _result.Add(KeyCode.KeypadMinus);

            _result.Add(KeyCode.F1);
            _result.Add(KeyCode.F2);
            _result.Add(KeyCode.F3);
            _result.Add(KeyCode.F4);

            _result.Add(KeyCode.F5);
            _result.Add(KeyCode.F6);
            _result.Add(KeyCode.F7);
            _result.Add(KeyCode.F8);

            _result.Add(KeyCode.F9);
            _result.Add(KeyCode.F10);
            _result.Add(KeyCode.F11);
            _result.Add(KeyCode.F12);

            _result.Add(KeyCode.Mouse0);
            _result.Add(KeyCode.Mouse1);
            _result.Add(KeyCode.Mouse2);
            _result.Add(KeyCode.Mouse3);
            _result.Add(KeyCode.Mouse4);
            _result.Add(KeyCode.Mouse5);

            return _result;
        }   

        public KeyCode GetKeyCodeFromName(string _keyName)
        {
            for(int i = 0; i < keyNames.Count; i++)
            {
                if(_keyName == keyNames[i])
                {
                    return keyCodes[i];
                }
            }

            return KeyCode.None;
        }

        private bool CheckForKeybindName(string _name)
        {
            foreach(KeyBindTrigger _t in triggers)
            {
                if (_t.ActionName.val == _name)
                    return true;
            }

            return false;
        }

        /* Callbacks */

        protected void Callback_GenerateTriggerIDList()
        {
            List<string> _list = new List<string>();

            for (int i = 0; i < triggers.Count; i++)
            {
                _list.Add(triggers[i].ActionName.val);
            }

            triggerChooser.choices = _list;
        }

        protected void Callback_OnTriggerSelected(string _name)
        {
            foreach(KeyBindTrigger _t in triggers)
            {
                if(_t.ActionName.val == _name)
                {
                    currentTrigger = _t;

                    triggerChooser.val = "Select Trigger";

                    UpdateKeyBindUI();

                    return;
                }
            }
        }

        protected void Callback_CreateNewTrigger()
        {
            if(CheckForKeybindName(nameInputField.UItext.text))
            {
                SuperController.LogError("You can't give the same name to multiple keybinds!!!");

                return;
            }

            triggers.Add(new KeyBindTrigger(this, triggers.Count));
            currentTrigger = triggers[triggers.Count - 1];

            triggerCount.val = triggerCount.val + 1;

            currentTrigger.ActionName.val = nameInputField.UItext.text;

            UpdateKeyBindUI();
        }

        protected void Callback_DestroyCurrentTrigger()
        {
            if(currentTrigger != null)
            {
                for(int i = 0; i < triggers.Count; i++)
                {
                    if(triggers[i] == currentTrigger)
                    {
                        currentTrigger.OnDestroy();

                        triggers.RemoveAt(i);

                        for (int j = 0; j < triggers.Count; j++)
                        {
                            triggers[j].UpdateID(j);
                        }

                        triggerCount.val = triggerCount.val - 1;

                        if (triggerCount.val > 0)
                            currentTrigger = triggers[triggers.Count - 1];
                        else
                            currentTrigger = null;

                        UpdateKeyBindUI();

                        return;
                    }
                }
            }
        }

        protected void Callback_OpenTriggerPanel()
        {
            if (currentTrigger == null)
                return;

            currentTrigger.Trigger.OpenPanelActionStart();
        }
   
        protected void Callback_RenameTrigger()
        {
            if (currentTrigger == null)
                return;

            if (CheckForKeybindName(nameInputField.UItext.text))
            {
                SuperController.LogError("You can't give the same name to multiple keybinds!!!");

                return;
            }

            currentTrigger.ActionName.val = nameInputField.UItext.text;

            UpdateKeyBindUI();
        }

        protected void Callback_OnKeySelected(string _keyName)
        {
            if (currentTrigger == null)
                return;

            currentTrigger.KeyID = _keyName;
        }

        /* Overrides */

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            if (includePhysical || forceStore)
            {
                needsStore = true;
                for (int i = 0; i < triggers.Count; i++)
                {
                    jc[triggers[i].Trigger.Name] = triggers[i].Trigger.GetJSON(base.subScenePrefix);
                }

            }
            return jc;
        }

        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);

            for (int i = 0; i < triggerCount.val; i++)
            {
                triggers.Add(new KeyBindTrigger(this, i));
            }

            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);

            if (triggerCount.val > 0)
                currentTrigger = triggers[0];

            UpdateKeyBindUI();
        }

        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
            {
                for(int i = 0; i < triggers.Count; i++)
                {
                    triggers[i].Trigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
                }
            }
        }

        /* Ui Helpers */

        protected List<UIDynamicSlider> activeSliders;
        protected List<UIDynamicToggle> activeToggles;
        protected List<UIDynamicButton> activeButtons;
        protected List<UIDynamicPopup> activePopups;
        protected List<UIDynamicTextField> activeTextFields;
        protected List<UIDynamic> activeSpacers;

        public void CreateTempSlider(JSONStorableFloat _data)
        {
            activeSliders.Add(CreateSlider(_data, true));
        }

        public void CreateTempToggle(JSONStorableBool _data)
        {
            activeToggles.Add(CreateToggle(_data, true));
        }

        public void CreateTempButton(string _text, UnityEngine.Events.UnityAction _callback)
        {
            UIDynamicButton _btn = CreateButton(_text, true);
            _btn.button.onClick.AddListener(_callback);

            activeButtons.Add(_btn);
        }

        public void CreateTempSpacer(float _height)
        {
            UIDynamic _spacer = CreateSpacer(true);
            _spacer.height = _height;
            activeSpacers.Add(_spacer);
        }

        public void CreateTempPopup(JSONStorableStringChooser _data)
        {
            activePopups.Add(CreatePopup(_data, true));
        }

        public void CreateTempFilterablePopup(JSONStorableStringChooser _data)
        {
            activePopups.Add(CreateFilterablePopup(_data, true));
        }

        public UIDynamicTextField CreateStaticText(string DescTitle, string DescText, bool rightSide, int fieldHeight)
        {
            JSONStorableString staticDescString = new JSONStorableString(DescTitle, DescText);
            UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
            staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
            LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
            sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
            staticDescStringField.height = fieldHeight;
            DisableScrollOnText(staticDescStringField);

            return staticDescStringField;
        }

        public UIDynamicTextField CreateTempStaticText(string DescTitle, string DescText, int fieldHeight)
        {
            activeTextFields.Add(CreateStaticText(DescTitle, DescText, true, fieldHeight));

            return activeTextFields[activeTextFields.Count - 1];
        }

        public UIDynamicTextField CreateTextInput(JSONStorableString jss, int fieldHeight, bool rightSide)
        {
            var textfield = CreateTextField(jss, rightSide);
            
            jss.dynamicText = textfield;

            textfield.backgroundColor = Color.white;

            var input = textfield.gameObject.AddComponent<InputField>();
            input.textComponent = textfield.UItext;
            jss.inputField = input;

            var layout = textfield.GetComponent<LayoutElement>();
            layout.preferredHeight = layout.minHeight = fieldHeight;

            textfield.height = fieldHeight;
            textfield.UItext.fontSize = 37;

            var rect = textfield.GetComponent<RectTransform>();
            rect.anchorMin = new Vector2(0, 1);
            rect.anchorMax = new Vector2(1, 1);
            rect.pivot = new Vector2(0, 0.5f);
            rect.anchoredPosition = new Vector2(0, -30f);
            rect.sizeDelta = new Vector2(0, 40f);

            return textfield;
        }

        private static void DisableScrollOnText(UIDynamicTextField target)
        {
            ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
            if (targetSR != null)
            {
                targetSR.horizontal = false;
                targetSR.vertical = false;
            }
        }

        public void ClearUI()
        {
            for (int i = 0; i < activeSliders.Count; i++)
            {
                RemoveSlider(activeSliders[i]);
            }

            for (int i = 0; i < activeToggles.Count; i++)
            {
                RemoveToggle(activeToggles[i]);
            }

            for (int i = 0; i < activeButtons.Count; i++)
            {
                RemoveButton(activeButtons[i]);
            }

            for (int i = 0; i < activeSpacers.Count; i++)
            {
                RemoveSpacer(activeSpacers[i]);
            }

            for (int i = 0; i < activePopups.Count; i++)
            {
                RemovePopup(activePopups[i]);
            }

            for (int i = 0; i < activeTextFields.Count; i++)
            {
                RemoveTextField(activeTextFields[i]);
            }

            activeTextFields = new List<UIDynamicTextField>();
            activeSpacers = new List<UIDynamic>();
            activeButtons = new List<UIDynamicButton>();
            activeSliders = new List<UIDynamicSlider>();
            activePopups = new List<UIDynamicPopup>();
            activeToggles = new List<UIDynamicToggle>();
        }
    }

    public class KeyBindTrigger
    {
        protected SimpleKeybind manager;
        protected string id;

        protected JSONStorableString js_Key_id;
        protected KeyCode keyCode;

        protected JSONStorableString js_ActionName;

        protected JSONStorableBool js_ShiftMod;
        protected JSONStorableBool js_CtrlMod;
        protected JSONStorableBool js_AltMod;

        protected EventTrigger trigger;

        public KeyBindTrigger(SimpleKeybind _manager, int _id)
        {
            manager = _manager;
            id = _id.ToString();

            js_Key_id = new JSONStorableString(id + "_KeyID", "");
            js_Key_id.setCallbackFunction += OnKeyIDChange;

            js_ActionName = new JSONStorableString(id + "_ActionName", "DefaultActionName");

            js_ShiftMod = new JSONStorableBool(id + "_ShiftMod", false);
            js_CtrlMod = new JSONStorableBool(id + "_CtrlMod", false);
            js_AltMod = new JSONStorableBool(id + "_AltMod", false);

            manager.RegisterString(js_ActionName);
            manager.RegisterString(js_Key_id);
            manager.RegisterBool(js_ShiftMod);
            manager.RegisterBool(js_CtrlMod);
            manager.RegisterBool(js_AltMod);

            trigger = new EventTrigger(manager, id + "_Trigger");
        }

        public void UpdateID(int _newID)
        {
            manager.DeregisterString(js_ActionName);
            manager.DeregisterString(js_Key_id);
            manager.DeregisterBool(js_ShiftMod);
            manager.DeregisterBool(js_CtrlMod);
            manager.DeregisterBool(js_AltMod);

            id = _newID.ToString();

            js_ActionName.name = id + "_ActionName";
            js_Key_id.name = id + "_KeyID";
            js_ShiftMod.name = id + "_ShiftMod";
            js_CtrlMod.name = id + "_CtrlMod";
            js_AltMod.name = id + "_AltMod";

            trigger.Name = id + "_Trigger";

            manager.RegisterString(js_ActionName);
            manager.RegisterString(js_Key_id);
            manager.RegisterBool(js_ShiftMod);
            manager.RegisterBool(js_CtrlMod);
            manager.RegisterBool(js_AltMod);
        }

        public void OnDestroy()
        {
            trigger.Remove();
        }

        public void OnKeyIDChange(string _newVal)
        {
            keyCode = manager.GetKeyCodeFromName(js_Key_id.val);
        }

        public string ID
        {
            get { return id; }
        }

        public string KeyID
        {
            set
            { js_Key_id.val = value; }
            get { return js_Key_id.val; }
        }

        public KeyCode Key
        {
            get { return keyCode; }
        }

        public JSONStorableString ActionName
        {
            get { return js_ActionName; }
        }

        public JSONStorableBool ShiftMod
        {
            get { return js_ShiftMod; }
        }

        public JSONStorableBool CtrlMod
        {
            get { return js_CtrlMod; }
        }

        public JSONStorableBool AltMod
        {
            get { return js_AltMod; }
        }

        public EventTrigger Trigger
        {
            get { return trigger; }
        }
    }

    public class SimpleTriggerHandler : TriggerHandler
    {
        public static bool Loaded { get; private set; }

        private static SimpleTriggerHandler myInstance;

        private RectTransform myTriggerActionsPrefab;
        private RectTransform myTriggerActionMiniPrefab;
        private RectTransform myTriggerActionDiscretePrefab;
        private RectTransform myTriggerActionTransitionPrefab;

        public static SimpleTriggerHandler Instance
        {
            get
            {
                if (myInstance == null)
                    myInstance = new SimpleTriggerHandler();
                return myInstance;
            }
        }

        public static void LoadAssets()
        {
            SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
        }

        private IEnumerator LoadAssetsInternal()
        {
            foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
                yield return x;

            Loaded = true;
        }

        private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
        {
            AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
            if (request == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
            yield return request;
            GameObject go = request.GetAsset<GameObject>();
            if (go == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
            RectTransform prefab = go.GetComponent<RectTransform>();
            if (prefab == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
            assign(prefab);
        }


        void TriggerHandler.RemoveTrigger(Trigger t)
        {
            // nothing to do
        }

        void TriggerHandler.DuplicateTrigger(Trigger t)
        {
            throw new NotImplementedException();
        }

        RectTransform TriggerHandler.CreateTriggerActionsUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionMiniUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
        }

        void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
        {
            UnityEngine.Object.Destroy(rt?.gameObject);
        }
    }
    // Wrapper for easier handling of custom triggers.
    public class EventTrigger : Trigger
    {
        public string Name
        {
            get { return name; }
            set { name = value; myNeedInit = true; }
        }

        public MVRScript Owner
        {
            get; private set;
        }

        private string name;
        private bool myNeedInit = true;

        public EventTrigger(MVRScript owner, string name)
        {
            Name = name;
            Owner = owner;
            handler = SimpleTriggerHandler.Instance;
        }

        public void OpenPanelActionStart()
        {
            if (!SimpleTriggerHandler.Loaded)
            {
                SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
                return;
            }

            triggerActionsParent = Owner.UITransform;
            InitTriggerUI();
            OpenTriggerActionsPanel();
            if (myNeedInit)
            {
                triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
                triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
                triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
                triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
                myNeedInit = false;
            }
        }

        public void OpenPanelTransition()
        {
            if (!SimpleTriggerHandler.Loaded)
            {
                SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
                return;
            }

            triggerActionsParent = Owner.UITransform;
            InitTriggerUI();
            OpenTriggerActionsPanel();

            //getRoot(triggerActionsParent,0);
            if (myNeedInit)
            {
                triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
                triggerActionsPanel.Find("Content/Tab2/Label").GetComponent<Text>().text = "Breathing Actions";
                triggerActionsPanel.Find("Content/Tab1").gameObject.SetActive(false);
                triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(true);
                triggerActionsPanel.Find("Content/Tab2").GetComponent<Toggle>().isOn = true;
                triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
                myNeedInit = false;
            }
        }

        public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
        {
            if (jc.HasKey(Name))
            {
                JSONClass tc = jc[Name].AsObject;
                if (tc != null)
                    base.RestoreFromJSON(tc, subScenePrefix, isMerge);
            }
            else if (setMissingToDefault)
            {
                base.RestoreFromJSON(new JSONClass());
            }
        }

        public void Trigger()
        {
            active = true;
            active = false;
        }

        public void TriggerTransition(float val)
        {
            this.transitionInterpValue = val;
        }
    }
}